#include "mpi.h"
#include <stdio.h>
#include <math.h>

/* This is the function to be integrated. */

double f( double a )
{
  return 4.0 / (1.0 + a*a);
}

/* This routine carries out integration of the 
function between limits x1 and x2 and it subdivides 
the interval into m parts. */

double integ( int m, double x1, double x2)
{
  int i;
  double sum=0.0, h, x;

  /* h is the step size. */

  h = (x2 - x1) / ((double)m);
  
  for( i=0 ; i<m ; i++ )
    { 
      x = x1 + h * ((double)(i-1));
      sum += f(x);
    }
  sum *= h;
  return(sum);
}

/* This is the main procedure.  It subdivides the range 
   over which the function is to be integrated into equal 
   parts.  The number of steps n is taken as an input from 
   the user on node 0, and communicated to all the other 
   nodes.  Local values of the integral are combined using 
   MPI_Reduce.  Output is printed on the console on node 0. */

int main( int argc, char *argv[])
{
  int n, m;
  int myid, numprocs;
  double x_low, x_up;
  double sum, x1, x2, dx;
  double mysum;
  char buffer[20];
  int position=0;

  /* Initialise MPI.  Get the number of processors and 
     rank of the local process. */

  MPI_Init(&argc,&argv);
  MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
  MPI_Comm_rank(MPI_COMM_WORLD,&myid);

  if (myid == 0)
    {
      printf("Enter the number of intervals: ");
      scanf("%d",&n);
      printf("Enter the lower limit of integration: (e.g., 0.0) ");
      scanf("%lf",&x_low);
      printf("Enter the upper limit of integration: (e.g., 1.0) ");
      scanf("%lf",&x_up);
    }
  
  MPI_Pack(&n,1,MPI_INT,&buffer,20,&position,MPI_COMM_WORLD);
  MPI_Pack(&x_low,1,MPI_DOUBLE,&buffer,20,&position,MPI_COMM_WORLD);
  MPI_Pack(&x_up,1,MPI_DOUBLE,&buffer,20,&position,MPI_COMM_WORLD);

  /* Broadcast the value of n, the number of 
     intervals to all other nodes.  I.e., all 
     nodes with myid not equal to zero. */
  
  MPI_Bcast(&buffer, 20, MPI_PACKED, 0, MPI_COMM_WORLD);

  position=0;

  MPI_Unpack(&buffer,20,&position,&n,1,MPI_INT,MPI_COMM_WORLD);
  MPI_Unpack(&buffer,20,&position,&x_low,1,MPI_DOUBLE,MPI_COMM_WORLD);
  MPI_Unpack(&buffer,20,&position,&x_up,1,MPI_DOUBLE,MPI_COMM_WORLD);

  /* Divide the range of integration into equal parts, 
     one for each processor. */
  
  dx = (x_up - x_low) / ((double) numprocs);
  
  /* Quit if the number of intervals is zero. */
  
  if (n != 0)
    {

      /* Limits of integration are computed on 
	 EACH node.  Each process computes the 
	 integral for a different section of 
	 the full range.  This is achieved by 
	 using myid, the rank of each process. */
      
      x1 = dx * ((double) myid);
      x2 = x1 + dx;
      m = n / numprocs;
      mysum = integ(m,x1,x2);
      
      /* Results of integration are collated 
	 from all the nodes. */
      
      MPI_Reduce(&mysum, &sum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
      
      /* And the result is printed on the 
	 screen on the 0th node. */
      
      if (myid == 0)
	{
	  printf("The value of the integral is %f\n",sum);
	  printf("\n");
	}
    }

  /* Finalise MPI */
  
  MPI_Finalize();
  return 0;
}

            
